/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.common.listener;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;

/**
 * This class helps a Component to listen MessageExchange on the DeliveryChannel
 * (accept() method). It acts as a Listener. When the listener calls the
 * 'accept()' method, its thread is blocked, but not the object that
 * instantiated it. The listener calls a 'process(Exchange)' on the
 * MessageExchangeProcessor, which acts as a handler of MessageExchanges. This
 * MessageExchangeProcessor has to be provided. <p/
 * <p>
 * The listener can be parametrized with a few parameters. See constructors.
 * </p>
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis
 */
public class MessageExchangeListener extends Thread {

	public enum IgnoredStatus {
		NOTHING_IGNORED, DONE_IGNORED, ERROR_IGNORED, DONE_AND_ERROR_IGNORED
	};

	private DeliveryChannel channel;

	private boolean running = false;

	private IgnoredStatus ignoredStatus;

	private long period;

	private MessageExchangeProcessor processor;

	private Logger logger; // NOPMD by gblondelle

	private LinkedBlockingQueue<Runnable> workers;

	private ThreadPoolExecutor executor;

	/**
	 * creates a listener on the specified channel. all exchanges are treated.
	 * no interval between two 'accept()';
	 * 
	 * @param channel
	 * @param processor
	 */
	public MessageExchangeListener(DeliveryChannel channel,
			MessageExchangeProcessor processor) {
		this(channel, processor, IgnoredStatus.NOTHING_IGNORED);
	}

	/**
	 * creates a listener on the specified channel. exchanges are ignored
	 * depending on the specified exchange-status to ignore. no interval between
	 * two 'accept()';
	 * 
	 * @param channel
	 * @param processor
	 * @param ignoredStatus
	 */
	public MessageExchangeListener(DeliveryChannel channel,
			MessageExchangeProcessor processor, IgnoredStatus ignoredStatus) {
		this(channel, processor, ignoredStatus, 0);
	}

	/**
	 * creates a listener on the specified channel. exchanges are ignored
	 * depending on the specified exchange-status to ignore. wait for the
	 * specified interval between two 'accept()';
	 * 
	 * @param channel
	 * @param processor
	 * @param ignoredStatus
	 * @param period
	 */
	public MessageExchangeListener(DeliveryChannel channel,
			MessageExchangeProcessor processor, IgnoredStatus ignoredStatus,
			long period) {
		this(channel, processor, IgnoredStatus.NOTHING_IGNORED, 0, "", null);
	}

	/**
	 * creates a listener on the specified channel. exchanges are ignored
	 * depending on the specified exchange-status to ignore. wait for the
	 * specified interval between two 'accept()';
	 * 
	 * @param channel
	 * @param processor
	 * @param ignoredStatus
	 * @param componentName
	 *            name of the component. Used to define the Thread name
	 * @param period
	 */
	public MessageExchangeListener(DeliveryChannel channel,
			MessageExchangeProcessor processor, IgnoredStatus ignoredStatus,
			long period, String componentName, Logger log) {
		super(componentName + "-JBI listener thread");
		this.ignoredStatus = ignoredStatus;
		this.channel = channel;
		this.period = period;
		this.processor = processor;
		this.logger = log;
		// TODO: The number of workers should be settable, example in a
		// properties file.
		this.workers = new LinkedBlockingQueue<Runnable>(50);
		// TODO: The following parameters should be settable, example in a
		// properties file:
		// - core pool size,
		// - maximum pool size,
		// - keep alive time
		this.executor = new ThreadPoolExecutor(10, 50, 600, TimeUnit.SECONDS,
				workers);
	}

	/**
	 * Close the listener. This cause the terminaison of the loop on
	 * DeliveryChannel.accept(), and interrupt this listening thread if the
	 * thread is blocked on DeliveryChannel.accept();
	 */
	public void close() {
		terminate();
		interrupt();
	}

	/**
	 * Start the thread listener ( Thread.start()). The listener waits on
	 * DeliveryChannel.accept(), treats the received MessageExchange, and waits
	 * again on DeliveryChannel.accept().
	 */
	public void listen() {
		start();
	}

	/**
	 * loop on accept
	 */
	public void run() {
		running = true;

		while (running) {
			MessageExchange exchange = null;

			// receive a message
			try {
				exchange = channel.accept();
			} catch (MessagingException e) {
				logger.log(Level.SEVERE, e.getMessage(), e);

			}

			// process the message
			if (!isIgnored(exchange)) {
				MessageExchangeWorker msgWorker = new MessageExchangeWorker(
						exchange, this.processor, this.logger);
				executor.execute(msgWorker);
			}

			// wait before next accept
			if (period > 0) {
				try {
					wait(period);

				} catch (InterruptedException e) {
					// nothing
				}
			}
		}
	}

	/**
	 * Terminate the listening. On the next MessageExchange treated after the
	 * DeliveryChannel.accept(), the listener will stop waiting on
	 * DeliveryChannel.accept().
	 */
	public void terminate() {
		running = false;

		// Stop workers that are working
		this.executor.shutdownNow();
	}

	/**
	 * depending on the IgnoredStatus, tells if the exchange has to be ignored.
	 * If the exchange is null, return true.
	 * 
	 * @param exchange
	 * @return
	 */
	protected boolean isIgnored(final MessageExchange exchange) {
		boolean result = false;

		if (exchange == null) {
			result = true;
		} else {
			ExchangeStatus status = exchange.getStatus();
			if (ExchangeStatus.DONE.equals(status)) {
				result = IgnoredStatus.DONE_IGNORED == ignoredStatus
						|| IgnoredStatus.DONE_AND_ERROR_IGNORED == ignoredStatus;
			} else if (ExchangeStatus.ERROR.equals(status)) {
				result = IgnoredStatus.ERROR_IGNORED == ignoredStatus
						|| IgnoredStatus.DONE_AND_ERROR_IGNORED == ignoredStatus;
			}
		}
		return result;
	}
}
